local strPackageName, nPackageVersion = "Dheva:AwesomePlate", 1

local AwesomePlate = {}

-- "Enum" for plate components
local Widgets = {
	["MAIN"] = 12,
	["HEALTHBAR"] = 1,
	["HEALTHBARBG"] = 11,
	["SHIELDBAR"] = 2,
	["HEALTHBARTEXT"] = 3,
	["HEALTHBARTEXTSTANDALONE"] = 10,
	["INTERRUPTARMOR"] = 4,
	["TARGETOUTLINE"] = 5,
	["RANKICON"] = 6,
	["CASTBAR"] = 7,
	["GUILD"] = 8,
	["QUEST"] = 9,
}

-- "Enum" for combat state
local CState = {
	["InCombat"] = true,
	["OutOfCombat"] = false
}

function AwesomePlate:new(o )
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	
	return o
end

function AwesomePlate:Init(tUnitP, tConfig, xmlDoc)
	self.tUnit = tUnitP
	
	self.wnds = {}

	self.wnds[Widgets.MAIN] = Apollo.LoadForm(xmlDoc, "Nameplate", "InWorldHudStratum", self)
	local wndMain = self.wnds[Widgets.MAIN]
	wndMain:Show(false, true)
	wndMain:SetUnit(tUnitP)
	
	self.wnds[Widgets.HEALTHBAR] = wndMain:FindChild("Healthbar")
	self.wnds[Widgets.HEALTHBARBG] = wndMain:FindChild("HealthbarBG")
	self.wnds[Widgets.SHIELDBAR] = wndMain:FindChild("Shieldbar")
	self.wnds[Widgets.HEALTHBARTEXT] = wndMain:FindChild("HealthBarText")
	self.wnds[Widgets.HEALTHBARTEXTSTANDALONE] = wndMain:FindChild("HealthBarTextStandAlone")
	self.wnds[Widgets.INTERRUPTARMOR] = wndMain:FindChild("IARFrame")
	self.wnds[Widgets.TARGETOUTLINE] = wndMain:FindChild("Outline")
	self.wnds[Widgets.RANKICON] = wndMain:FindChild("Icon")
	self.wnds[Widgets.CASTBAR] = wndMain:FindChild("Castbar")
	self.wnds[Widgets.GUILD] = wndMain:FindChild("Guild")
	self.wnds[Widgets.QUEST] = wndMain:FindChild("QuestFrame")
		
	self.nDisposition = tUnitP:GetDispositionTo(GameLib.GetPlayerUnit())
	self.tUnitConfig, self.tConfig = FindConfig(tUnitP, tConfig)
	self.tGeneralConfig = tConfig
	
	self:UpdateCombatState() -- Calls InitAll anyway
	
	-- SET STATUS FLAGS
	self.bOccluded = self.wnds[Widgets.MAIN]:IsOccluded()
	self.bOnScreen = self.wnds[Widgets.MAIN]:IsOnScreen() 	
	self.bInRange = self:CheckRange()
	
	self.bDestroyed = false
	
	self:UpdateVisibility()
end

function AwesomePlate:Destroy()
	for key, wnd in pairs(self.wnds) do
		wnd:Destroy()
	end
end

function AwesomePlate:InitAll()
	self:InitBarColors()
	self:InitHealth()
	self:InitShield()
	self:InitInterruptArmor()
	self:InitName()
	self:InitAffiliation()
	self:InitTarget()
	self:InitReward()
	self:InitCastState()
	self:InitRankIcon()
end

function AwesomePlate:UpdateCombatState()
	_, self.tConfig = FindConfig(self.tUnit, self.tGeneralConfig)
	self:InitAll()
end

function AwesomePlate:InitAffiliation()
	if self.tConfig["bAffiliation"] then
		local strAffiliation = self.tUnit:GetAffiliationName()
		if strAffiliation ~= nil then
			if strAffiliation ~= "" then
				if self.tUnit:GetType() == "Player" then
					strAffiliation = String_GetWeaselString(Apollo.GetString("Nameplates_GuildDisplay"), strAffiliation)
				end
			end
		    self.wnds[Widgets.GUILD]:SetText(strAffiliation)
			self.wnds[Widgets.GUILD]:Show(true, true)
		else
			self.wnds[Widgets.GUILD]:Show(false, true)		
		end	
	else
		self.wnds[Widgets.GUILD]:Show(false, true)
	end
end

function AwesomePlate:InitBarColors()
	self.wnds[Widgets.HEALTHBARTEXT]:SetTextColor(self.tConfig["HealthtextColor"])
	self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:SetTextColor(self.tConfig["HealthtextSOColor"])
	local cTextColor = self.tConfig["HealthtextSOColor"]
	if self.tConfig["bHealthbar"] then
		cTextColor = self.tConfig["HealthtextColor"]
	end
	self.wnds[Widgets.GUILD]:SetTextColor(cTextColor)
    self.wnds[Widgets.HEALTHBAR]:SetBarColor(self.tConfig["HealthbarColor"])
	self.wnds[Widgets.SHIELDBAR]:SetBarColor(self.tConfig["ShieldbarColor"])
	--Print("Setting bg color for unit "..self.tUnit:GetName().." to: "..self.tConfig["HealthbarColor"].r.." "..self.tConfig["HealthbarColor"].b.." "..self.tConfig["HealthbarColor"].g.." "..self.tConfig["HealthbarColor"].a)
end

function AwesomePlate:InitHealth()
	if self.tConfig['bHealthbar'] then
		local nHealth = self.tUnit:GetHealth()
		local nMaxHealth = self.tUnit:GetMaxHealth()
		if nHealth == nil or nMaxHealth == nil then
			nHealth = 1
			nMaxHealth = 1
		end
		self.wnds[Widgets.HEALTHBAR]:SetMax(nMaxHealth)
		self.wnds[Widgets.HEALTHBAR]:SetProgress(nHealth)
		self.wnds[Widgets.HEALTHBAR]:Show(true, true)
		self.wnds[Widgets.HEALTHBARBG]:Show(true, true)
	else
		self.wnds[Widgets.HEALTHBAR]:Show(false, true)
		self.wnds[Widgets.SHIELDBAR]:Show(false, true)
		self.wnds[Widgets.HEALTHBARTEXT]:Show(false, true)
		self.wnds[Widgets.HEALTHBARBG]:Show(false, true)
		self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:Show(self.tConfig["bName"], true)
	end	
end

function AwesomePlate:UpdateHealth()
	local nHealth = self.tUnit:GetHealth()
	local nMaxHealth = self.tUnit:GetMaxHealth()
	if nHealth == nil or nMaxHealth == nil then
		nHealth = 1
		nMaxHealth = 1
	end
	self.wnds[Widgets.HEALTHBAR]:SetMax(nMaxHealth)
	self.wnds[Widgets.HEALTHBAR]:SetProgress(nHealth)
	if nHealth == 0 then
		self:Destroy()
		self.bDestroyed = true -- Flag used to delete this object later
	end
end

function AwesomePlate:InitShield()
	if self.tConfig["bHealthbar"] then
		local nShield = self.tUnit:GetShieldCapacity()
		local nMaxShield = self.tUnit:GetShieldCapacityMax()
		self.wnds[Widgets.SHIELDBAR]:SetMax(nMaxShield or 1)
		self.wnds[Widgets.SHIELDBAR]:SetProgress(nShield or 0)
		--self.wnds[Widgets.SHIELDBAR]:SetBarColor(self.tUnitConfig.ShieldbarColor)
		self.wnds[Widgets.SHIELDBAR]:Show(true, true)
	else
		self.wnds[Widgets.SHIELDBAR]:Show(false, true)
	end	
end

function AwesomePlate:UpdateShield()
	local nShield = self.tUnit:GetShieldCapacity()
	local nMaxShield = self.tUnit:GetShieldCapacityMax()
	self.wnds[Widgets.SHIELDBAR]:SetMax(nMaxShield or 1)
	self.wnds[Widgets.SHIELDBAR]:SetProgress(nShield or 0)
end

function AwesomePlate:InitInterruptArmor()
	local nIAR = self.tUnit:GetInterruptArmorValue()
	if self.tConfig["bInterruptArmor"] and nIAR > 0 then
		self.wnds[Widgets.INTERRUPTARMOR]:SetText(nIAR)
		self.wnds[Widgets.INTERRUPTARMOR]:Show(true, true)
	else
		self.wnds[Widgets.INTERRUPTARMOR]:Show(false, true)
	end	
end

function AwesomePlate:UpdateInterruptArmor()
	local nIAR = self.tUnit:GetInterruptArmorValue()
	if nIAR > 0 then
		self.wnds[Widgets.INTERRUPTARMOR]:SetText(nIAR)
		self.wnds[Widgets.INTERRUPTARMOR]:Show(true, true)
	else
		self.wnds[Widgets.INTERRUPTARMOR]:Show(false, true)
	end	
end

function AwesomePlate:InitTarget()
	if self.tConfig["bTarget"] then
		local bIsTarget = (GameLib.GetTargetUnit() == self.tUnit)
		self.wnds[Widgets.TARGETOUTLINE]:Show(bIsTarget, true)
	else
		self.wnds[Widgets.TARGETOUTLINE]:Show(false, true)
	end
end

function AwesomePlate:UpdateTarget()
	local bIsTarget = (self.tConfig["bTarget"] and (GameLib.GetTargetUnit() == self.tUnit))
	self.wnds[Widgets.TARGETOUTLINE]:Show(bIsTarget, true)
end

function AwesomePlate:InitName()
	local strName = self.tUnit:GetName()
	if self.tConfig["bName"] then
		if self.tConfig["bHealthbar"] then
			self.wnds[Widgets.HEALTHBARTEXT]:SetText(strName)
			self.wnds[Widgets.HEALTHBARTEXT]:Show(true,true)
			self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:Show(false,true)
		else
			self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:SetText(strName)
			self.wnds[Widgets.HEALTHBARTEXT]:Show(false,true)
			self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:Show(true,true)
		end
	else
		self.wnds[Widgets.HEALTHBARTEXT]:Show(false,true)
		self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:Show(false,true)
	end	
end

function AwesomePlate:UpdateName()
	local strName = self.tUnit:GetName()
	self.wnds[Widgets.HEALTHBARTEXT]:SetText(strName)
	self.wnds[Widgets.HEALTHBARTEXTSTANDALONE]:SetText(strName)
end

function AwesomePlate:InitCastState()
	if self.tConfig["bCastBar"] then
		if self.tUnit:ShouldShowCastBar() then
			-- Unit is casting
			self.wnds[Widgets.CASTBAR]:SetFillSprite("CRB_TargetFrameSprites:sprTF_CastMeterRed")
			self.wnds[Widgets.CASTBAR]:SetProgress(self.tUnit:GetCastTotalPercent())
			self.wnds[Widgets.CASTBAR]:SetMax(100)
			self.wnds[Widgets.CASTBAR]:SetText(self.tUnit:GetCastName())
			self.wnds[Widgets.CASTBAR]:Show(true, true)
		else
			local fMooTime = self.tUnit:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Vulnerability)
			if fMooTime > 0.0 then
				-- Unit is vulnerable
				if self.fMooDuration == nil then
					-- newly detected
					self.fMooDuration = fMooTime
					self.wnds[Widgets.HEALTHBAR]:SetBarColor(self.tGeneralConfig.MooColor)
					self.wnds[Widgets.CASTBAR]:SetFillSprite("CRB_TargetFrameSprites:sprTF_CastMeterOrange")
					self.wnds[Widgets.CASTBAR]:SetMax(fMooTime)
					self.wnds[Widgets.CASTBAR]:SetProgress(fMooTime)
					self.wnds[Widgets.CASTBAR]:SetText("Vulnerability")
					self.wnds[Widgets.CASTBAR]:Show(true, true)
				else
					-- already detected before
					self.wnds[Widgets.CASTBAR]:SetProgress(fMooTime)
				end
			else
				self.fMooDuration = nil
				self.wnds[Widgets.HEALTHBAR]:SetBarColor(self.tConfig["HealthbarColor"])
				self.wnds[Widgets.CASTBAR]:Show(false, true)
			end
		end
	else
		self.wnds[Widgets.CASTBAR]:Show(false, true)
	end						
end

function AwesomePlate:UpdateCastState()
	self:InitCastState() -- TODO: some things might not be necessary				
end

function AwesomePlate:InitReward()
	if self.tConfig["bQuest"] then
		local tFlags = {
			bVert = false,
			bHideQuests = false,
			bHideChallenges = false,
			bHideMissions = false,
			bHidePublicEvents = false,
			bHideRivals = true,
			bHideFriends = true
		}
		if RewardIcons ~= nil and RewardIcons.GetUnitRewardIconsForm ~= nil then
			RewardIcons.GetUnitRewardIconsForm(self.wnds[Widgets.QUEST], self.tUnit, tFlags)
			self.wnds[Widgets.QUEST]:Show(true, true)
		else
			self.wnds[Widgets.QUEST]:Show(false, true)
		end
	else
		self.wnds[Widgets.QUEST]:Show(false, true)	
	end	
end

function AwesomePlate:UpdateReward()
	local tFlags = {
		bVert = false,
		bHideQuests = false,
		bHideChallenges = false,
		bHideMissions = false,
		bHidePublicEvents = false,
		bHideRivals = true,
		bHideFriends = true
	}
	if RewardIcons ~= nil and RewardIcons.GetUnitRewardIconsForm ~= nil then
		RewardIcons.GetUnitRewardIconsForm(self.wnds[Widgets.QUEST], self.tUnit, tFlags)
		self.wnds[Widgets.QUEST]:Show(true, true)
	else
		self.wnds[Widgets.QUEST]:Show(false, true)	
	end
end

function AwesomePlate:InitRankIcon()
	if self.tConfig["bRankIcon"] then
		if self.tUnit:IsACharacter() then
			-- Player character, set class icon
			local strSprite = self.tGeneralConfig["ClassIconSprites"][self.tUnit:GetClassId()].Sprite
			local tOffsets = self.tGeneralConfig["ClassIconOffsets"]
			local fScale = self.tGeneralConfig["ClassIconScale"]
			self.wnds[Widgets.RANKICON]:SetSprite(strSprite)
			self.wnds[Widgets.RANKICON]:SetAnchorOffsets(tOffsets[1] , tOffsets[2] , tOffsets[3] , tOffsets[4])
			self.wnds[Widgets.RANKICON]:SetScale(fScale)
			self.wnds[Widgets.RANKICON]:Show(true, true)
		else
			-- NPC, set difficulty icon (those are not of equal size, so adjust offsets and scaling)
			if self.tGeneralConfig["RankIconSprites"][self.tUnit:GetRank()] ~= nil then
				local strSprite = self.tGeneralConfig["RankIconSprites"][self.tUnit:GetRank()].Sprite
				--Print("Setting sprite to "..strSprite)
				local tOffset = self.tGeneralConfig["RankIconSprites"][self.tUnit:GetRank()].Offsets
				local fScale = self.tGeneralConfig["RankIconSprites"][self.tUnit:GetRank()].Scale
				self.wnds[Widgets.RANKICON]:SetSprite(strSprite)
				self.wnds[Widgets.RANKICON]:SetAnchorOffsets(tOffset[1], tOffset[2], tOffset[3], tOffset[4])
				self.wnds[Widgets.RANKICON]:SetScale(fScale)
				self.wnds[Widgets.RANKICON]:Show(true, true)
			else
				self.wnds[Widgets.RANKICON]:Show(false, true)
			end
		end
	else
		self.wnds[Widgets.RANKICON]:Show(false, true)
	end
end

function AwesomePlate:UpdateDisposition()
	self.tUnitConfig, self.tConfig = FindConfig(self.tUnit, self.tGeneralConfig)
	self:InitAll()
end

function AwesomePlate:OnTenthSecTimer()
	if self.wnds[Widgets.MAIN]:IsShown() then
		if self.tConfig["bHealthbar"] then
			self:UpdateHealth()
			self:UpdateShield()
		end
		if self.tConfig["bCastBar"] then
			self:UpdateCastState()
		end
	end
end

function AwesomePlate:OnHalfSecTimer()
	self.bInRange = self:CheckRange()

	if self.wnds[Widgets.MAIN]:IsShown() then
		if self.tConfig["bInterruptArmor"] then
			self:UpdateInterruptArmor()
		end
	end
	
	if self.nDisposition ~= self.tUnit:GetDispositionTo(GameLib.GetPlayerUnit()) then
		self.nDisposition = self.tUnit:GetDispositionTo(GameLib.GetPlayerUnit())
		self:UpdateDisposition()
	end	
	
	self:UpdateVisibility()
end

function AwesomePlate:OnNameplateClicked(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)	
	if GameLib.GetTargetUnit() ~= self.tUnit and eMouseButton == GameLib.CodeEnumInputMouse.Left then
		GameLib.SetTargetUnit(self.tUnit)
	end
end

function AwesomePlate:OnUnitOcclusionChanged(wndHandler, wndControl, bOccluded)
	self.bOccluded = bOccluded 
	self:UpdateVisibility()
end

function AwesomePlate:OnWorldLocationOnScreen(wndHandler, wndControl, bOnScreen)
	self.bOnScreen = bOnScreen
	self:UpdateVisibility()
end

function AwesomePlate:UpdateVisibility()
	local bInstant = false
	if self.bInRange and self.bOnScreen and not(self.bOccluded) then
		self.wnds[Widgets.MAIN]:Show(true, bInstant)
	else
		self.wnds[Widgets.MAIN]:Show(false, bInstant)
	end
end

function AwesomePlate:CheckRange()
	if self.tUnit:IsThePlayer() then
		return true
	end
	local fSquareDist = CalcSquareDist(GameLib.GetPlayerUnit(), self.tUnit)
	return (fSquareDist <= 70*70)  -- TODO config value
end

Apollo.RegisterPackage(AwesomePlate, strPackageName, nPackageVersion , {})
-----------------------------------------------------------------------------
-- WINDOW DEFS
-----------------------------------------------------------------------------


-----------------------------------------------------------------------------
-- HELPER FUNCTIONS
-----------------------------------------------------------------------------
function CalcSquareDist(tUnit, tUnit2)
	local tFirstPos = tUnit:GetPosition()
	local tSecPos = tUnit2:GetPosition()
	
	if tFirstPos == nil or tSecPos == nil then
		return math.huge
	end
	
	local dx = tSecPos.x - tFirstPos.x
	local dy = tSecPos.y - tFirstPos.y
	local dz = tSecPos.z - tFirstPos.z
	
	local fSquareDist = dx*dx + dy*dy + dz*dz
	
	return fSquareDist
end

function FindConfig(tUnit, tConfig)
	local tMinion = nil
	-- If pet, use owner's settings
	if tUnit:GetUnitOwner() ~= nil then
		tMinion = tUnit
		tUnit = tUnit:GetUnitOwner()
	end
	
	if tUnit:IsThePlayer() then
		strCategory = "player"	
	elseif tUnit:IsACharacter() then
		local nDispo = tUnit:GetDispositionTo(GameLib.GetPlayerUnit())
		if nDispo == Unit.CodeEnumDisposition.Hostile then
			strCategory = "hostilePlayer"		
		else
			strCategory = "friendlyPlayer"
		end
	else
		local nDispo = tUnit:GetDispositionTo(GameLib.GetPlayerUnit())
		if nDispo == Unit.CodeEnumDisposition.Hostile then
			strCategory = "hostileNPC"		
		elseif nDispo == Unit.CodeEnumDisposition.Neutral then
			strCategory = "neutralNPC"
		else
			strCategory = "friendlyNPC"
		end
	end
	
	local strCState = "OutOfCombat"
	if GameLib.GetPlayerUnit():IsInCombat() then
		strCState = "InCombat"
	end
		
	return tConfig[strCategory], tConfig[strCategory][strCState]
end