-----------------------------------------------------------------------------------------------
-- Client Lua Script for PlatesOfAwesome
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "Unit"
require "GameLib"
require "Apollo"
require "ApolloTimer"
require "math"
require "string"
 
-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome Module Definition
-----------------------------------------------------------------------------------------------
local PlatesOfAwesome = {} 

local GeminiGUI = nil

local AwesomePlate = nil

local GeminiColor = nil
 
-----------------------------------------------------------------------------------------------
-- Constants<
-----------------------------------------------------------------------------------------------
local karStaticConfig = {
	["MooColor"] = ApolloColor.new(1.0, 0.0, 1.0):ToTable(),
	["RankIconSprites"] = {
		[Unit.CodeEnumRank.Champion] = {Sprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Champion", Offsets = {-25, -8, 0, 22}, Scale = 0.9},
		[Unit.CodeEnumRank.Elite] = {Sprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Elite", Offsets = {-25, -4, 7, 36}, Scale = 0.75},
		[Unit.CodeEnumRank.Fodder] = {Sprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Fodder", Offsets = {-25, -12, 0, 22}, Scale = 1.2},
		[Unit.CodeEnumRank.Minion] = {Sprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Minion", Offsets = {-25, -11, 0, 22}, Scale = 1.1},
		[Unit.CodeEnumRank.Standard] = {Sprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Standard", Offsets = {-25, -10, 0, 22}, Scale = 1.0},
		[Unit.CodeEnumRank.Superior] = {Sprite = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Superior", Offsets = {-25, -7, 0, 32}, Scale = 0.8},
	},
	["ClassIconSprites"] = {
		[GameLib.CodeEnumClass.Engineer] = {Sprite = "IconSprites:Icon_Windows_UI_CRB_Engineer"},
		[GameLib.CodeEnumClass.Esper] = {Sprite = "IconSprites:Icon_Windows_UI_CRB_Esper"},
		[GameLib.CodeEnumClass.Medic] = {Sprite = "IconSprites:Icon_Windows_UI_CRB_Medic"},
		[GameLib.CodeEnumClass.Spellslinger] = {Sprite = "IconSprites:Icon_Windows_UI_CRB_Spellslinger"},
		[GameLib.CodeEnumClass.Stalker] = {Sprite = "IconSprites:Icon_Windows_UI_CRB_Stalker"},
		[GameLib.CodeEnumClass.Warrior] = {Sprite = "IconSprites:Icon_Windows_UI_CRB_Warrior"},
	},
	["ClassIconOffsets"] = {-25, -3, 0, 22},
	["ClassIconScale"] = 1.0,
	["player"] = {
		["InCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
		["OutOfCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
	},
	["friendlyNPC"] = {
		
		
		["InCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
		["OutOfCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
	},
	["neutralNPC"] = {	
		["InCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
		["OutOfCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
	},
	["hostileNPC"] = {
		["InCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
		["OutOfCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
	},
	["friendlyPlayer"] = {
		["InCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
		["OutOfCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
	},
	["hostilePlayer"] = {	
		["InCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
		["OutOfCombat"] = {
			["bShieldbar"] = true,
			["bHidden"] = false,
		},
	}
}

local SettingTypes = {
	["BOOL"] = 0,
	["COLOR"] = 1,
}

local karConfigAnnotations = {
	["player"] = {
			["strName"] = "Player",
			
			["HealthbarColor"] = {["strText"] = "Healthbar color (fill)", ["type"] = SettingTypes.COLOR},
			["ShieldbarColor"] = {["strText"] = "Shieldbar color (fill)", ["type"] = SettingTypes.COLOR},
			["HealthtextColor"] = {["strText"] = "Healthtext color", ["type"] = SettingTypes.COLOR},
			["HealthtextSOColor"] = {["strText"] = "Name color (Healthbar faded)", ["type"] = SettingTypes.COLOR},
			
			["bPets"] = {["strText"] = "Show pets", ["type"] = SettingTypes.BOOL},
			["bHealthbar"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bTarget"] = {["strText"] = "Show target outline", ["type"] = SettingTypes.BOOL},
			["bQuest"] = {["strText"] = "Show quest indicator", ["type"] = SettingTypes.BOOL},
			["bName"] = {["strText"] = "Show name", ["type"] = SettingTypes.BOOL},
			["bAffiliation"] = {["strText"] = "Show guild name", ["type"] = SettingTypes.BOOL},
			["bRankIcon"] = {["strText"] = "Show class icon", ["type"] = SettingTypes.BOOL},
			["bInterruptArmor"] = {["strText"] = "Show interrupt armor", ["type"] = SettingTypes.BOOL},
			["bCastBar"] = {["strText"] = "Show castbar", ["type"] = SettingTypes.BOOL},
	},
	["friendlyNPC"] = {
			["strName"] = "Friendly NPCs",
			
			["HealthbarColor"] = {["strText"] = "Healthbar color (fill)", ["type"] = SettingTypes.COLOR},
			["ShieldbarColor"] = {["strText"] = "Shieldbar color (fill)", ["type"] = SettingTypes.COLOR},
			["HealthtextColor"] = {["strText"] = "Healthtext color", ["type"] = SettingTypes.COLOR},
			["HealthtextSOColor"] = {["strText"] = "Name color (Healthbar faded)", ["type"] = SettingTypes.COLOR},
			
			["bPets"] = {["strText"] = "Show pets", ["type"] = SettingTypes.BOOL},
			["bHealthbar"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bTarget"] = {["strText"] = "Show target outline", ["type"] = SettingTypes.BOOL},
			["bQuest"] = {["strText"] = "Show quest indicator", ["type"] = SettingTypes.BOOL},
			["bName"] = {["strText"] = "Show name", ["type"] = SettingTypes.BOOL},
			["bAffiliation"] = {["strText"] = "Show guild name", ["type"] = SettingTypes.BOOL},
			["bRankIcon"] = {["strText"] = "Show class icon", ["type"] = SettingTypes.BOOL},
			["bInterruptArmor"] = {["strText"] = "Show interrupt armor", ["type"] = SettingTypes.BOOL},
			["bCastBar"] = {["strText"] = "Show castbar", ["type"] = SettingTypes.BOOL},
	},	
	["neutralNPC"] = {
			["strName"] = "Neutral NPCs",
			
			["HealthbarColor"] = {["strText"] = "Healthbar color (fill)", ["type"] = SettingTypes.COLOR},
			["ShieldbarColor"] = {["strText"] = "Shieldbar color (fill)", ["type"] = SettingTypes.COLOR},
			["HealthtextColor"] = {["strText"] = "Healthtext color", ["type"] = SettingTypes.COLOR},
			["HealthtextSOColor"] = {["strText"] = "Name color (Healthbar faded)", ["type"] = SettingTypes.COLOR},
			
			["bPets"] = {["strText"] = "Show pets", ["type"] = SettingTypes.BOOL},
			["bHealthbar"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bTarget"] = {["strText"] = "Show target outline", ["type"] = SettingTypes.BOOL},
			["bQuest"] = {["strText"] = "Show quest indicator", ["type"] = SettingTypes.BOOL},
			["bName"] = {["strText"] = "Show name", ["type"] = SettingTypes.BOOL},
			["bAffiliation"] = {["strText"] = "Show guild name", ["type"] = SettingTypes.BOOL},
			["bRankIcon"] = {["strText"] = "Show class icon", ["type"] = SettingTypes.BOOL},
			["bInterruptArmor"] = {["strText"] = "Show interrupt armor", ["type"] = SettingTypes.BOOL},
			["bCastBar"] = {["strText"] = "Show castbar", ["type"] = SettingTypes.BOOL},
	},	
	["hostileNPC"] = {
			["strName"] = "Hostile NPCs",
			
			["HealthbarColor"] = {["strText"] = "Healthbar color (fill)", ["type"] = SettingTypes.COLOR},
			["ShieldbarColor"] = {["strText"] = "Shieldbar color (fill)", ["type"] = SettingTypes.COLOR},
			["HealthtextColor"] = {["strText"] = "Healthtext color", ["type"] = SettingTypes.COLOR},
			["HealthtextSOColor"] = {["strText"] = "Name color (Healthbar faded)", ["type"] = SettingTypes.COLOR},
			
			["bPets"] = {["strText"] = "Show pets", ["type"] = SettingTypes.BOOL},
			["bHealthbar"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bTarget"] = {["strText"] = "Show target outline", ["type"] = SettingTypes.BOOL},
			["bQuest"] = {["strText"] = "Show quest indicator", ["type"] = SettingTypes.BOOL},
			["bName"] = {["strText"] = "Show name", ["type"] = SettingTypes.BOOL},
			["bAffiliation"] = {["strText"] = "Show guild name", ["type"] = SettingTypes.BOOL},
			["bRankIcon"] = {["strText"] = "Show class icon", ["type"] = SettingTypes.BOOL},
			["bInterruptArmor"] = {["strText"] = "Show interrupt armor", ["type"] = SettingTypes.BOOL},
			["bCastBar"] = {["strText"] = "Show castbar", ["type"] = SettingTypes.BOOL},
	},	
	["friendlyPlayer"] = {
			["strName"] = "Friendly Players",
			
			["HealthbarColor"] = {["strText"] = "Healthbar color (fill)", ["type"] = SettingTypes.COLOR},
			["ShieldbarColor"] = {["strText"] = "Shieldbar color (fill)", ["type"] = SettingTypes.COLOR},
			["HealthtextColor"] = {["strText"] = "Healthtext color", ["type"] = SettingTypes.COLOR},
			["HealthtextSOColor"] = {["strText"] = "Name color (Healthbar faded)", ["type"] = SettingTypes.COLOR},
			
			["bPets"] = {["strText"] = "Show pets", ["type"] = SettingTypes.BOOL},
			["bHealthbar"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bTarget"] = {["strText"] = "Show target outline", ["type"] = SettingTypes.BOOL},
			["bQuest"] = {["strText"] = "Show quest indicator", ["type"] = SettingTypes.BOOL},
			["bName"] = {["strText"] = "Show name", ["type"] = SettingTypes.BOOL},
			["bAffiliation"] = {["strText"] = "Show guild name", ["type"] = SettingTypes.BOOL},
			["bRankIcon"] = {["strText"] = "Show class icon", ["type"] = SettingTypes.BOOL},
			["bInterruptArmor"] = {["strText"] = "Show interrupt armor", ["type"] = SettingTypes.BOOL},
			["bCastBar"] = {["strText"] = "Show castbar", ["type"] = SettingTypes.BOOL},
	},	
	["hostilePlayer"] = {
			["strName"] = "Hostile Players",
			
			["HealthbarColor"] = {["strText"] = "Healthbar color (fill)", ["type"] = SettingTypes.COLOR},
			["ShieldbarColor"] = {["strText"] = "Shieldbar color (fill)", ["type"] = SettingTypes.COLOR},
			["HealthtextColor"] = {["strText"] = "Healthtext color", ["type"] = SettingTypes.COLOR},
			["HealthtextSOColor"] = {["strText"] = "Name color (Healthbar faded)", ["type"] = SettingTypes.COLOR},
			
			["bPets"] = {["strText"] = "Show pets", ["type"] = SettingTypes.BOOL},
			["bNonFlaggedAreHostile"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bHealthbar"] = {["strText"] = "Show healthbar", ["type"] = SettingTypes.BOOL},
			["bTarget"] = {["strText"] = "Show target outline", ["type"] = SettingTypes.BOOL},
			["bQuest"] = {["strText"] = "Show quest indicator", ["type"] = SettingTypes.BOOL},
			["bName"] = {["strText"] = "Show name", ["type"] = SettingTypes.BOOL},
			["bAffiliation"] = {["strText"] = "Show guild name", ["type"] = SettingTypes.BOOL},
			["bRankIcon"] = {["strText"] = "Show class icon", ["type"] = SettingTypes.BOOL},
			["bInterruptArmor"] = {["strText"] = "Show interrupt armor", ["type"] = SettingTypes.BOOL},
			["bCastBar"] = {["strText"] = "Show castbar", ["type"] = SettingTypes.BOOL},
	}
}

local karDefaultConfig = {
	["player"] = {
		["InCombat"] = {
			["HealthbarColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = false,
			["bAffiliation"] = false,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		},
		["OutOfCombat"] = {
			["HealthbarColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = false,
			["bAffiliation"] = false,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		}
	},
	["friendlyNPC"] = {
		["InCombat"] = {
			["HealthbarColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = false,
			["bAffiliation"] = false,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		},
		["OutOfCombat"] = {
			["HealthbarColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		}
	},
	["neutralNPC"] = {
		["InCombat"] = {
			["HealthbarColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = true,
			["bHealthbar"] = true,
			["bTarget"] = true,
			["bQuest"] = true,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = true,
			["bInterruptArmor"] = true,
			["bCastBar"] = true,
		},
		["OutOfCombat"] = {
			["HealthbarColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = true,
			["bHealthbar"] = true,
			["bTarget"] = true,
			["bQuest"] = true,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = true,
			["bInterruptArmor"] = true,
			["bCastBar"] = true,
		}
	},
	["hostileNPC"] = {
		["InCombat"] = {
			["HealthbarColor"] = ApolloColor.new(1.0, 0.0, 0.0):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = true,
			["bHealthbar"] = true,
			["bTarget"] = true,
			["bQuest"] = true,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = true,
			["bInterruptArmor"] = true,
			["bCastBar"] = true,
		},
		["OutOfCombat"] = {
			["HealthbarColor"] = ApolloColor.new(1.0, 0.0, 0.0):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = true,
			["bHealthbar"] = true,
			["bTarget"] = true,
			["bQuest"] = true,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = true,
			["bInterruptArmor"] = true,
			["bCastBar"] = true,
		}
	},
	["friendlyPlayer"] = {
		["InCombat"] = {
			["HealthbarColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		},
		["OutOfCombat"] = {
			["HealthbarColor"] = ApolloColor.new("xkcdApple"):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		}
	},
	["hostilePlayer"] = {
		["InCombat"] = {
			["HealthbarColor"] = ApolloColor.new(1.0, 0.0, 0.0):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		},
		["OutOfCombat"] = {
			["HealthbarColor"] = ApolloColor.new(1.0, 0.0, 0.0):ToTable(),
			["ShieldbarColor"] = ApolloColor.new("xkcdBluish"):ToTable(),
			["HealthtextColor"] = ApolloColor.new("white"):ToTable(),
			["HealthtextSOColor"] = ApolloColor.new(1.0, 1.0, 0.0):ToTable(),
			["bPets"] = false,
			["bHealthbar"] = false,
			["bTarget"] = false,
			["bQuest"] = false,
			["bName"] = true,
			["bAffiliation"] = true,
			["bRankIcon"] = false,
			["bInterruptArmor"] = false,
			["bCastBar"] = false,
		}
	}
}

-- "Enum" for plate components
local Widgets = {
	["MAIN"] = 12,
	["HEALTHBAR"] = 1,
	["HEALTHBARBG"] = 11,
	["SHIELDBAR"] = 2,
	["HEALTHBARTEXT"] = 3,
	["HEALTHBARTEXTSTANDALONE"] = 10,
	["INTERRUPTARMOR"] = 4,
	["TARGETOUTLINE"] = 5,
	["RANKICON"] = 6,
	["CASTBAR"] = 7,
	["GUILD"] = 8,
	["QUEST"] = 9,
}

-- "Enum" for combat state
local CState = {
	["InCombat"] = true,
	["OutOfCombat"] = false
}

local tiPreloadTimer = nil

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function PlatesOfAwesome:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
		
	self.arPreloadUnits = {}
	self.nLastTargetId = nil	
    return o
end

function PlatesOfAwesome:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Plates Of Awesome"
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
	-- Try to load units until all units have been loaded
end
 

-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome OnLoad
-----------------------------------------------------------------------------------------------
function PlatesOfAwesome:OnLoad()
	Apollo.RegisterEventHandler("UnitCreated", "OnPreloadUnitCreated", self)

	self.bXmlDocLoaded = false
	self.bPlayerUnitLoaded = false
	self.bSpritesLoaded = false
	self.bConfigLoaded = false
	
	self.xmlDoc = XmlDoc.CreateFromFile("PlatesOfAwesome.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	Apollo.LoadSprites("POASprites.xml", "POASprites")
	
end

-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome OnDocLoaded
-----------------------------------------------------------------------------------------------
function PlatesOfAwesome:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.bXmlDocLoaded = true
		self.nPreloadRuns = 0
		tiPreloadTimer = ApolloTimer.Create(2, true, "OnPreloadTimer", self)
		tiPreloadTimer:Start()
	end
end

function PlatesOfAwesome:OnPreloadTimer()
	self.nPreloadRuns = self.nPreloadRuns + 1
	
	self.bPlayerUnitLoaded = (GameLib.GetPlayerUnit() ~= nil)
	self.bSpritesLoaded = Apollo.IsSpriteLoaded("HealthFill") and Apollo.IsSpriteLoaded("HealthFillShaded")
	
	if self.bPlayerUnitLoaded and self.bSpritesLoaded and self.bXmlDocLoaded and self.bConfigLoaded then
		Apollo.RemoveEventHandler("UnitCreated", self)
		Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)

		tiPreloadTimer:Stop()
		self:StartUp()
	end
		
	if self.bPlayerUnitLoaded and self.bSpritesLoaded and self.bXmlDocLoaded and self.nPreloadRuns == 3 then
		Apollo.RemoveEventHandler("UnitCreated", self)
		Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)

		self:OnRestore(GameLib.CodeEnumAddonSaveLevel.Character, nil)
		
		tiPreloadTimer:Stop()
		self:StartUp()
	end
end

function PlatesOfAwesome:StartUp()
	self.arUnit2Nameplate = {}
	
	Apollo.RegisterSlashCommand("poa", "OnConfigure", self)
	Apollo.RegisterSlashCommand("awesome", "OnConfigure", self)
	Apollo.RegisterSlashCommand("poadebug", "OnDebug", self)
	
	GeminiGUI = Apollo.GetPackage("Gemini:GUI-1.0").tPackage
	AwesomePlate = Apollo.GetPackage("Dheva:AwesomePlate").tPackage
	GeminiColor = Apollo.GetPackage("GeminiColor").tPackage

	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	--Apollo.RegisterEventHandler("NextFrame", "OnNextFrame", self) Removed to avoid fps impact
	Apollo.RegisterTimerHandler("halfSecTimer", "OnHalfSecTimer", self)
	Apollo.CreateTimer("halfSecTimer", 0.5, true)
	Apollo.RegisterTimerHandler("tenthSecTimer", "OnTenthSecTimer", self)
	Apollo.CreateTimer("tenthSecTimer", 0.1, true)
	Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnitChanged", self)
	Apollo.RegisterEventHandler("UnitNameChanged", "OnUnitNameChanged", self)
	--Apollo.RegisterEventHandler("UnitGibbed", "OnUnitGibbed", self) Removed, see handler function
	Apollo.RegisterEventHandler("UnitTextBubbleCreate", "OnUnitTextBubbleCreate", self) 
	Apollo.RegisterEventHandler("UnitTextBubblesDestroyed", "OnUnitTextBubblesDestroyed", self) 
	Apollo.RegisterEventHandler("UnitEnteredCombat", "OnUnitEnteredCombat", self) 
	Apollo.RegisterEventHandler("UnitTitleChanged", "OnUnitTitleChanged", self)
	Apollo.RegisterEventHandler("PlayerTitleChange", "OnPlayerTitleChange", self) 
	Apollo.RegisterEventHandler("UnitGuildNameplateChanged", "OnUnitGuildNameplateChanged", self) 
	Apollo.RegisterEventHandler("UnitLevelChanged", "OnUnitLevelChanged", self) 
	Apollo.RegisterEventHandler("UnitMemberOfGuildChange", "OnUnitMemberOfGuildChange", self) 
	Apollo.RegisterEventHandler("GuildChange", "OnGuildChange", self)
	Apollo.RegisterEventHandler("ChangeWorld", "OnChangeWorld", self)
	--Apollo.RegisterEventHandler("RefreshHealthShieldBar", "OnRefreshHealthShieldBar", self) Removed, see handler function
	
	local tRewardUpdateEvents = {
		"QuestObjectiveUpdated", "QuestStateChanged", "ChallengeAbandon", "ChallengeLeftArea",
		"ChallengeFailTime", "ChallengeFailArea", "ChallengeActivate", "ChallengeCompleted",
		"ChallengeFailGeneric", "PublicEventObjectiveUpdate", "PublicEventUnitUpdate",
		"PlayerPathMissionUpdate", "FriendshipAdd", "FriendshipPostRemove", "FriendshipUpdate" 
	}
	for idx, value in pairs(tRewardUpdateEvents) do
		Apollo.RegisterEventHandler(value, "OnRewardUpdate", self)
	end
	
	for id, tUnit in pairs(self.arPreloadUnits) do
		self:OnUnitCreated(tUnit)
		self.arPreloadUnits[id] = nil
	end

end

-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome EVENT HANDLER
-----------------------------------------------------------------------------------------------

function PlatesOfAwesome:OnPreloadUnitCreated(unitNew)
	self.arPreloadUnits[unitNew:GetId()] = unitNew
end

function PlatesOfAwesome:OnChangeWorld()
	for key, tNameplate in pairs(self.arUnit2Nameplate) do
		tNameplate:Destroy()
		self.arUnit2Nameplate[key] = nil
	end
	self.nPreloadRuns = 0
	tiPreloadTimer = ApolloTimer.Create(2, true, "OnPreloadTimer", self)
end

function PlatesOfAwesome:OnUnitCreated(tUnit)
	if GameLib.GetPlayerUnit() == nil then
		self.arPreloadUnits[tUnit:GetId()] = tUnit
		return
	end

	if not self:ShouldCreateNameplate(tUnit) then
		return
	end
	
	local tNameplate = AwesomePlate:new(nil)
	tNameplate:Init(tUnit, self.tConfig, self.xmlDoc)
	self.arUnit2Nameplate[tNameplate.tUnit:GetId()] = tNameplate
	--Apollo.RegisterEventHandler("UnitOcclusionChanged", "OnUnitOcclusionChanged", tNameplate.wnds[CState.OutOfCombat][Widgets.MAIN])
end

function PlatesOfAwesome:OnUnitDestroyed(tUnit)
	local tNameplate = self.arUnit2Nameplate[tUnit:GetId()]
	if tNameplate ~= nil then
		tNameplate:Destroy()
		self.arUnit2Nameplate[tUnit:GetId()] = nil
	end
end

function PlatesOfAwesome:OnHalfSecTimer()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		if tNameplate.bDestroyed then
			self.arUnit2Nameplate[idx] = nil
		end
		tNameplate:OnHalfSecTimer()
	end
end

function PlatesOfAwesome:OnTenthSecTimer()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		if tNameplate.bDestroyed == true then
			self.arUnit2Nameplate[idx] = nil
		end
		tNameplate:OnTenthSecTimer()
	end
end

function PlatesOfAwesome:OnGuildChange()
	
end

function PlatesOfAwesome:OnUnitMemberOfGuildChange()
	
end

function PlatesOfAwesome:OnUnitLevelChanged()
	
end

function PlatesOfAwesome:OnUnitGuildNameplateChanged(tUnit)
	if tUnit ~= nil then
		local tNameplate = self.arUnit2Nameplate[tUnit:GetId()]
		--self:UpdateAffiliation(tNameplate)
		tNameplate:InitAffiliation()
	end	
end

function PlatesOfAwesome:OnPlayerTitleChange()
	
end

function PlatesOfAwesome:OnUnitTitleChanged()
	
end

function PlatesOfAwesome:OnUnitEnteredCombat(tUnit)
	if tUnit ~= nil and tUnit:IsThePlayer() then
		for key, tNameplate in pairs(self.arUnit2Nameplate) do
			tNameplate:UpdateCombatState()
		end
	end
end

function PlatesOfAwesome:OnUnitNameChanged(tUnit)
	if tUnit ~= nil and self.arUnit2Nameplate[tUnit:GetId()] ~= nil then
		self.arUnit2Nameplate[tUnit:GetId()]:UpdateName()
	end
end

function PlatesOfAwesome:OnTargetUnitChanged(uNewTarget)
	if uNewTarget ~= nil then
		local tNameplate = self.arUnit2Nameplate[uNewTarget:GetId()]
		if tNameplate ~= nil then
			tNameplate:UpdateTarget()
		end
	end
	
	if self.nLastTargetId ~= nil then
		local tNameplate = self.arUnit2Nameplate[self.nLastTargetId]
		if tNameplate ~= nil then
			tNameplate:UpdateTarget()
		end
	end
	
	if uNewTarget ~= nil then
		self.nLastTargetId = uNewTarget:GetId()
	else
		self.nLastTargetId = nil
	end
end

function PlatesOfAwesome:OnRewardUpdate()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		tNameplate:UpdateReward()
	end
end

-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome functions
-----------------------------------------------------------------------------------------------

function PlatesOfAwesome:RecreateAll()
	for key, tNameplate in pairs(self.arUnit2Nameplate) do
		local tUnit = tNameplate.tUnit
		tNameplate:Destroy()
		self.arUnit2Nameplate[tUnit:GetId()] = nil
		self:OnUnitCreated(tUnit)
	end
end

function PlatesOfAwesome:ShouldCreateNameplate(tUnit)
	return self.arUnit2Nameplate[tUnit:GetId()] == nil and 
		(
			(
			tUnit:IsValid() 
			and tUnit:ShouldShowNamePlate() 
			and tUnit:GetType() ~= "Collectible" 
			and tUnit:GetType() ~= "PinataLoot" 
			and tUnit:GetType() ~= "Mount"
			and not tUnit:IsMyGhost()
			)
			--or (kbShowSettlerRessources and tUnit:GetAffiliation() == "<Settler Resource>")
		)
end

-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome Config
-----------------------------------------------------------------------------------------------

function PlatesOfAwesome:OnConfigure()	
	if self.wndOptions == nil then
		self:CreateOptionsWindow()
	end
	self:SyncDataToWindow(self.wndOptions)
	self.wndOptions:Show(true, false)
end

function PlatesOfAwesome:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end

	local tSave = karDefaultConfig
	self:MergeTables(self.tConfig, tSave, true)
	return tSave
end

function PlatesOfAwesome:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end

	self.tConfig = {}
	self:DeepCopyTable(karDefaultConfig, self.tConfig)
	
	if tSavedData ~= nil then
		-- Overwrite existing with user settings
		self:MergeTables(tSavedData, self.tConfig, true)
	end
	
	-- Add static settings
	self:AddTable(karStaticConfig, self.tConfig)
	
	self.bConfigLoaded = true
end

function PlatesOfAwesome:MergeTables(tSource, tTarget, bOverride)
	for key, value in pairs(tTarget) do
		if type(value) == "table" then
			self:MergeTables(tSource[key], tTarget[key], bOverride)
		else
			if (bOverride and tTarget[key] ~= nil) or not bOverride then
				tTarget[key] = tSource[key]
			end
		end
	end
end

function PlatesOfAwesome:AddTable(tSource, tTarget)
	for key, value in pairs(tSource) do
		if type(value) == "table" then
			if tTarget[key] == nil then
				tTarget[key] = {}
			end
			self:AddTable(tSource[key], tTarget[key])
		else
			tTarget[key] = tSource[key]
		end
	end
end

function PlatesOfAwesome:DeepCopyTable(tSource, tTarget)
	for key, value in pairs(tSource) do
		if type(value) == "table" then
			tTarget[key] = {}
			self:DeepCopyTable(tSource[key], tTarget[key])
		else
			tTarget[key] = tSource[key]
		end
	end
end

function PlatesOfAwesome:CreateOptionsWindow()
	local nNumCategories = 1 -- Add one for general settings
	for k,v in pairs(karDefaultConfig) do nNumCategories = nNumCategories + 1 end
	
	local nWindowWidth = 800
	local nWindowHeight = 600
	-- Create the window itself
	local tWindowDefinition = {
		Name         	= "OptionsWindow",
		Template      	= "Metal_Primary_Nav",
		UseTemplateBG 	= true,
		Picture       	= true,
		Border        	= true,
		Moveable		= true,
		AnchorCenter  	= { nWindowWidth, nWindowHeight },
		TextColor		= ApolloColor.new("White"),
		--VScroll			= true,
		Children		= {}
	}
	
	-- Title
	local tTitle = {
		Name			= "Title",
		Text			= "Plates Of Awesome Settings",
		Font			= "CRB_Header20",
		AnchorOffsets 	= {10, 30, nWindowWidth - 60, 60},
		DT_CENTER 		= true		
	}
	table.insert(tWindowDefinition.Children, tTitle)

	-- Create category buttons
	local nButtonWidth = math.floor((nWindowWidth - 2*26) / nNumCategories + 0.5)
	local nCurX = 26
	
	local tGeneralButton = {
		WidgetType		= "PushButton",
		Name		 	= "generalBtn",
		Base			= "BK3:btnMetal_TabMainLeft",
		Text		  	= "General",
        AnchorOffsets 	= {nCurX, 49, nCurX + nButtonWidth - 1, 125},
		Events			= { ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnOptionsCategoryButtonClicked(a,b,c) end }
	}
	table.insert(tWindowDefinition.Children, tGeneralButton)
	nCurX = nCurX + nButtonWidth
	
	for strCategory, tSettings in pairs(karDefaultConfig) do
		strBase = "BK3:btnMetal_TabMainMid"
		if nCurX >= (nNumCategories-1)*nButtonWidth+26 then
			strBase = "BK3:btnMetal_TabMainRight"
		end
		local tCatButton = {
			WidgetType		= "PushButton",
			Name		 	= strCategory.."Btn",
			Base			= strBase,
			Text		  	= karConfigAnnotations[strCategory]["strName"],
	        AnchorOffsets 	= {nCurX, 49, nCurX + nButtonWidth - 1, 125},
			Data			= strCategory,
			Events			= { ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnOptionsCategoryButtonClicked(a,b,c) end }
		}
		table.insert(tWindowDefinition.Children, tCatButton)
		nCurX = nCurX + nButtonWidth
	end
	
	-- Create Category Windows
	local tCatGeneralWindow = {
		Template		= "Holo_Background_General",
		Name			= "generalWnd",
		Picture			= true,
		AnchorOffsets	= {0, 140, nWindowWidth-60, nWindowHeight-22},
		VScroll			= true,
		Visible			= false,
		Children		= {},
		--Text			= "Welcome to Plates Of Awesome, the most awesome nameplates in the world!"
	}
	
	local tExplain0 = {
		Text = "Welcome to Plates Of Awesome!",
		Font = "CRB_Header14",
		AnchorOffsets = { 10, 10, nWindowWidth - 60, 100 }
	}
	
	local tExplain1 = {
		Text = "In this dialog, you can adjust PoA's settings to your needs.\n\nTo save your changes, you can use those three little buttons on the top right side: \n",
		AnchorOffsets = { 10, 50, nWindowWidth - 60, 100 }
	}
	
	local tExplain2 = {
		WidgetType    = "PushButton",
        AnchorOffsets = {8, 104, 40, 140},
        Base		  = "CRB_ChatLogSprites:btnCL_Save",
		Scale		  = 1.4,
	}
	local tExplain3 = {
		Text = "Apply settings",
		DT_VCENTER = true,
		AnchorOffsets = { 43, 100, nWindowWidth - 60, 135 },
		
	}
	
	local tExplain4 = {
		WidgetType    = "PushButton",
        AnchorOffsets = {10, 131, 40, 170},
        Base        = "CRB_QuestTrackerSprites:btnQT_QuestRedeem",
	}
	local tExplain5 = {
		Text = "Apply and close",
		DT_VCENTER = true,
		AnchorOffsets = { 43, 130, nWindowWidth - 60, 165 },
	}
	
	local tExplain6 = {
		WidgetType    = "PushButton",
        AnchorOffsets = {15, 169, 40, 200},
        Base          = "CRB_Basekit:kitBtn_Holo_Close",
	}
	local tExplain7 = {
		Text = "Close without saving",
		DT_VCENTER = true,
		AnchorOffsets = { 43, 160, nWindowWidth - 60, 200 },
	}
	
	local tExplain8 = {
		Text = "Created by Dheva - GUI is using GeminiGUI and GeminiColor",
		DT_VCENTER = true,
		AnchorOffsets = { 10, 400, nWindowWidth - 60, 430 },
	}
	table.insert(tCatGeneralWindow.Children, tExplain0)
	table.insert(tCatGeneralWindow.Children, tExplain1)
	table.insert(tCatGeneralWindow.Children, tExplain2)
	table.insert(tCatGeneralWindow.Children, tExplain3)
	table.insert(tCatGeneralWindow.Children, tExplain4)
	table.insert(tCatGeneralWindow.Children, tExplain5)
	table.insert(tCatGeneralWindow.Children, tExplain6)
	table.insert(tCatGeneralWindow.Children, tExplain7)
	table.insert(tCatGeneralWindow.Children, tExplain8)
	table.insert(tWindowDefinition.Children, tCatGeneralWindow)
	
	for strCategory, tSettings in pairs(karDefaultConfig) do
		local tCatWindow = {
			Template		= "Holo_Background_General",
			Name			= strCategory.."Wnd",
			Picture			= true,
			AnchorOffsets	= {0, 140, nWindowWidth-60, nWindowHeight-22},
			VScroll			= true,
			Visible			= false,
			Children		= {}
		}
		
		local nCatWinWidth = nWindowWidth-60
		
		-- VISIBILITY BOOLEANS
		local tTitles = {
			Name			= strCategory.."TitleRow",
			AnchorOffsets	= {10, 10, nCatWinWidth/2, 45},
			
			Children		= {
				{Name = "Col2", Text = "In\nCombat", AnchorOffsets = {0, 0, 45, 35}, Font = "CRB_Header9", DT_VCENTER = true},
				{Name = "Col3", Text = "Out Of\nCombat", AnchorOffsets = {45, 0, 100, 35}, Font = "CRB_Header9", DT_VCENTER = true},
				{Name = "Col1", Text = "Visibility", AnchorOffsets = {90, 0, 500, 35}, Font = "CRB_Header14", DT_VCENTER = true},
			},
		}
		
		local tOptions = {
			Name			= strCategory.."Options",
			AnchorOffsets	= {10, 45, nCatWinWidth/2,nWindowHeight-22-45},
			Children		= {}
		}
		local nCurY = 2 
		local nLineHeight = 22
		for strName, tSetting in pairs(karDefaultConfig[strCategory]["InCombat"]) do
			if karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.BOOL then
				local strControlName = strCategory.."/"..strName
				
				local tName = {
					Name	= strCategory.."/"..strName,
					Text	= karConfigAnnotations[strCategory][strName]["strText"],
					AnchorOffsets	= {90,nCurY+5,500,nCurY+nLineHeight}, 
				}
				table.insert(tOptions.Children, tName)
				
				local tIC = {
					Name			= strControlName.."/IC",
					WidgetType		= "CheckBox",
					AnchorOffsets	= {5,nCurY,45,nCurY+nLineHeight}, 
				}
				table.insert(tOptions.Children, tIC)
				
				local tOOC = {
					Name			= strControlName.."/OOC",
					WidgetType		= "CheckBox",				
					AnchorOffsets	= {45,nCurY,90,nCurY+nLineHeight}, 
				}
				table.insert(tOptions.Children, tOOC)
				nCurY = nCurY + nLineHeight
			end
		end
		table.insert(tCatWindow.Children, tTitles)
		table.insert(tCatWindow.Children, tOptions)

		-- COLOR SETTINGS
		local xOffsetPos = nCatWinWidth/2
		local tTitlesCol = {
			Name			= strCategory.."TitleRowColor",
			AnchorOffsets	= {10+nCatWinWidth/2, 10, nWindowWidth-70, 45},
			
			Children		= {
				{Name = "Col2", Text = "In\nCombat", AnchorOffsets = {0, 0, 45, 35}, Font = "CRB_Header9", DT_VCENTER = true},
				{Name = "Col2", Text = "Out of\nCombat", AnchorOffsets = {45, 0, 90, 35}, Font = "CRB_Header9", DT_VCENTER = true},
				{Name = "Col1", Text = "Colors", AnchorOffsets = {90, 0, 500, 35}, Font = "CRB_Header14", DT_VCENTER = true},
			},
		}
		
		local tOptions = {
			Name			= strCategory.."OptionsColor",
			AnchorOffsets	= {10+nCatWinWidth/2, 45, nWindowWidth-70, nWindowHeight-22-45},
			Children		= {}
		}
		local nCurY = 2 
		local nLineHeight = 22
		for strName, tSetting in pairs(karDefaultConfig[strCategory]["InCombat"]) do
			if karConfigAnnotations[strCategory][strName] ~= nil then
				if karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.COLOR then
					local strControlName = strCategory.."/"..strName
					
					local tName = {
						Name	= strControlName,
						Text	= karConfigAnnotations[strCategory][strName]["strText"],
						AnchorOffsets	= {90,nCurY+3,500,nCurY+nLineHeight}, 
					}
					table.insert(tOptions.Children, tName)
					
					local tIC = {
						Name			= strControlName.."/Color/IC",
						WidgetType		= "Button",
						BGColor			= ApolloColor.new("white"),
						Sprite			= "BasicSprites:WhiteFill",
						Picture			= true,
						AnchorOffsets	= {0,nCurY+5,40,nCurY+nLineHeight-5},
						Events			= {
							ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnSelectColor(strControlName.."/Color/IC",a,b,c) end
						}
					}
					table.insert(tOptions.Children, tIC)
														
					local tOOC = {
						Name			= strControlName.."/Color/OOC",
						WidgetType		= "Button",
						BGColor			= ApolloColor.new("white"),
						Sprite			= "BasicSprites:WhiteFill",
						Picture			= true,
						AnchorOffsets	= {45,nCurY+5,85,nCurY+nLineHeight-5},
						Events			= {
							ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnSelectColor(strControlName.."/Color/OOC",a,b,c) end
						}
					}
					table.insert(tOptions.Children, tOOC)
				
					nCurY = nCurY + nLineHeight
				end
			end
			
		end
		table.insert(tCatWindow.Children, tTitlesCol)
		table.insert(tCatWindow.Children, tOptions)
		
		table.insert(tWindowDefinition.Children, tCatWindow)
	end
	
	-- Window control buttons
	local tOkButton = {
		WidgetType    = "PushButton",
		Name		  = "OkButton",
        AnchorOffsets = {nWindowWidth - 75, 23, nWindowWidth - 50, 48},
        Base          = "CRB_QuestTrackerSprites:btnQT_QuestRedeem",
        NoClip        = false,
		Events = { ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnOptionsOkButtonClicked(a,b,c) end }
	}
	table.insert(tWindowDefinition.Children, tOkButton)
	
	local tApplyButton = {
		WidgetType    = "PushButton",
		Name		  = "ApplyButton",
        AnchorOffsets = {nWindowWidth - 102, 23, nWindowWidth -77, 50},
        NoClip        = false,
		Base		  = "CRB_ChatLogSprites:btnCL_Save",
		Scale		  = 1.4,
		Events = { ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnOptionsApplyButtonClicked(a,b,c) end }	
	}
	table.insert(tWindowDefinition.Children, tApplyButton)

	local tCancelButton = {
		WidgetType    = "PushButton",
		Name		  = "CancelButton",
		Text		  = "",
        AnchorOffsets = {nWindowWidth - 50, 25, nWindowWidth - 25, 50},
        Base          = "CRB_Basekit:kitBtn_Holo_Close",
        NoClip        = false,
		Events = { ButtonSignal = function(a,b,c) Apollo.GetAddon("PlatesOfAwesome"):OnOptionsCancelButtonClicked(a,b,c) end }	
	}
	table.insert(tWindowDefinition.Children, tCancelButton)
	
	local wndInstance = GeminiGUI:Create(tWindowDefinition):GetInstance()
	self.wndOptions = wndInstance
	
	self:OnOptionsCategoryButtonClicked(nil, nil, wndInstance:FindChild("generalBtn"))
end

function PlatesOfAwesome:SyncWindowToData(wndInstance)
	for strCategory, tSettings in pairs(karDefaultConfig) do
		for strName, tSetting in pairs(tSettings["InCombat"]) do
			if karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.BOOL then
				local strControlName = strCategory.."/"..strName.."/OOC"
				local wndChildOOC = wndInstance:FindChild(strControlName)
				strControlName = strCategory.."/"..strName.."/IC"
				local wndChildIC = wndInstance:FindChild(strControlName)		
				self.tConfig[strCategory]["InCombat"][strName] = wndChildIC:IsChecked()
				self.tConfig[strCategory]["OutOfCombat"][strName] = wndChildOOC:IsChecked()
			
			elseif karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.COLOR then
				local strControlName = strCategory.."/"..strName.."/Color/IC"
				local wndChildIC = wndInstance:FindChild(strControlName)
				strControlName = strCategory.."/"..strName.."/Color/OOC"
				local wndChildOOC = wndInstance:FindChild(strControlName)
				local tColorOOC = wndChildOOC:GetBGColor():ToTable()
				local tColorIC = wndChildIC:GetBGColor():ToTable()
				self.tConfig[strCategory]["InCombat"][strName] = tColorIC
				self.tConfig[strCategory]["OutOfCombat"][strName] = tColorOOC
			end
		end
		for strName, tSetting in pairs(tSettings) do
			if karConfigAnnotations[strCategory][strName] ~= nil
			and karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.COLOR then
				local strControlName = strCategory.."/"..strName.."/Color"
				local wndChild = wndInstance:FindChild(strControlName)
				self.tConfig[strCategory][strName] = wndChild:GetBGColor():ToTable()
			end
		end
	end
end

function PlatesOfAwesome:SyncDataToWindow(wndInstance)
	for strCategory, tSettings in pairs(karDefaultConfig) do
		for strName, tSetting in pairs(tSettings["InCombat"]) do
			if karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.BOOL then
				local setting = self.tConfig[strCategory]["InCombat"][strName]
				local strControlName = strCategory.."/"..strName.."/IC"
				local wndChild = wndInstance:FindChild(strControlName)
				if wndChild == nil then
					Print("Cannot find control "..strControlName)
				end
				wndChild:SetCheck(setting)
				
				setting = self.tConfig[strCategory]["OutOfCombat"][strName]
				strControlName = strCategory.."/"..strName.."/OOC"
				wndChild = wndInstance:FindChild(strControlName)
				if wndChild == nil then
					Print("Cannot find control "..strControlName)
				end
				wndChild:SetCheck(setting)
			elseif karConfigAnnotations[strCategory][strName]["type"] == SettingTypes.COLOR then
				
				local tColor = self.tConfig[strCategory]["InCombat"][strName]
				local strControlName = strCategory.."/"..strName.."/Color/IC"
				local wndChild = wndInstance:FindChild(strControlName)
				wndChild:SetBGColor(tColor)
				
				tColor = self.tConfig[strCategory]["OutOfCombat"][strName]
				strControlName = strCategory.."/"..strName.."/Color/OOC"
				wndChild = wndInstance:FindChild(strControlName)
				wndChild:SetBGColor(tColor)
			else
				Print("Could not find value type for: "..strCategory.."/"..strName)
			end
		end
	end
end

function PlatesOfAwesome:OnOptionsCategoryButtonClicked(_, _, wndControl)
	local strNewCatWindowName = string.sub(wndControl:GetName(), 1, -4).."Wnd"
	local wndCurCat = self.wndCurCat
	if self.wndCurCat ~= nil then
		wndCurCat:Show(false)
		wndControl:GetParent():FindChild(string.sub(wndCurCat:GetName(),0,-4).."Btn"):SetCheck(false)
	end
	local wndNewCat = wndControl:GetParent():FindChild(strNewCatWindowName)
	wndNewCat:Show(true) 
	self.wndCurCat = wndNewCat
	wndControl:SetCheck(true)
end

function PlatesOfAwesome:OnOptionsOkButtonClicked(_, wndHandler, wndControl)
	self:SyncWindowToData(wndControl:GetParent())
	wndControl:GetParent():Close() 
	self:RecreateAll() 
end

function PlatesOfAwesome:OnOptionsApplyButtonClicked(_, wndHandler, wndControl)
	self:SyncWindowToData(wndControl:GetParent())
	self:RecreateAll()
end

function PlatesOfAwesome:OnOptionsCancelButtonClicked(_, wndHandler, wndControl)
	self.wndOptions:Show(false)
	self:SyncDataToWindow(wndControl:GetParent())
end

function PlatesOfAwesome:OnSelectColor(strCtrlName, a, b, c)
	local initColor = self.wndOptions:FindChild(strCtrlName):GetBGColor()
	if self.wndColor ~= nil then
		self.wndColor:Destroy()
	end
	self.wndColor = GeminiColor:CreateColorPicker(self, "OnColorPicker", false, initColor, strCtrlName)
end

function PlatesOfAwesome:OnColorPicker(strColor, strCtrlName)
	local wndChild = self.wndOptions:FindChild(strCtrlName)
	wndChild:SetBGColor(ApolloColor.new(strColor))
end

-----------------------------------------------------------------------------------------------
-- DEBUG FUNCTIONS
-----------------------------------------------------------------------------------------------

function PlatesOfAwesome:OnDebug()
	if self.wndDebug ~= nil then
		self:OnRefreshDebugWindow()
		self.wndDebug:Show(true, false)
		return
	end
	local tWindow = {
		Name         	= "DebugWindow",
		Template      	= "Holo_Large",
		--Sprite			= "BK3:spr_BK3_Holo_Alert_Confirm_Blocker",
		UseTemplateBG 	= true,
		Picture       	= true,
		Border        	= true,
		Moveable		= true,
		AnchorOffsets  	= { 0, 0, 400, 1000},
		TextColor		= ApolloColor.new("White"),
		--VScroll			= true,
		Children		= {}
	}
	
	
	local tNumPlates = {
		Name			= "NumPlatesDisplay",
		--Text			= "Active plates: "..num,
		AnchorOffsets 	= {10,10, 360, 25}
	}
	table.insert(tWindow.Children, tNumPlates)
	
	local tNumPlates = {
		Name			= "NumShownPlatesDisplay",
		--Text			= "Active plates: "..num,
		AnchorOffsets 	= {10,25, 360, 40}
	}
	table.insert(tWindow.Children, tNumPlates)
	
	local tUnitGrid = {
		Name			= "UnitGrid",
		WidgetType		= "Grid",
		AnchorOffsets 	= {10,40, 308, 600},
		Columns = {
			{ Name = "ID", Width = 80 },
			{ Name = "Name", Width = 308-10-80 },
        },
		VScroll			= true,
	}
	table.insert(tWindow.Children, tUnitGrid)
	
	local tExitButton = {
		WidgetType 			= "PushButton",
		Text				= "Close",
		AnchorOffsets		= {210, 920, 360, 950},
		Events				= { ButtonSignal = function() Apollo.GetAddon("PlatesOfAwesome"):OnExitDebugWindow() end }
	}
	table.insert(tWindow.Children, tExitButton)
	
	local tRefreshButton = {
		WidgetType 			= "PushButton",
		Text				= "Refresh",
		AnchorOffsets		= {40, 920, 190, 950},
		Events				= { ButtonSignal = function() Apollo.GetAddon("PlatesOfAwesome"):OnRefreshDebugWindow() end }
	}
	table.insert(tWindow.Children, tRefreshButton)
	
	local wndInstance = GeminiGUI:Create(tWindow):GetInstance()
	wndInstance:Show(true, false)
	self.wndDebug = wndInstance
	
	self:OnRefreshDebugWindow()
end

function PlatesOfAwesome:OnRefreshDebugWindow()
	local num = 0
	local active = 0
	for key, value in pairs(self.arUnit2Nameplate) do
		if value.wnds[Widgets.MAIN]:IsShown() then
			active = active + 1
		end
		num = num + 1
	end
	
	local wndNum = self.wndDebug:FindChild("NumPlatesDisplay")
	wndNum:SetText("Existing plates: "..num)
	
	local wndNum = self.wndDebug:FindChild("NumShownPlatesDisplay")
	wndNum:SetText("Shown plates: "..active)
	
	local wndUnitGrid = self.wndDebug:FindChild("UnitGrid")
	wndUnitGrid:DeleteAll()
	for key, tNameplate in pairs(self.arUnit2Nameplate) do
		local iCurrRow =  wndUnitGrid:AddRow("")
		wndUnitGrid:SetCellText(iCurrRow, 1, tNameplate.tUnit:GetId())
		wndUnitGrid:SetCellText(iCurrRow, 2, tNameplate.tUnit:GetName())
	end
end

function PlatesOfAwesome:OnExitDebugWindow()
	if self.wndDebug ~= nil then
		self.wndDebug:Destroy()
		self.wndDebug = nil
	end
end

-----------------------------------------------------------------------------------------------
-- NOTES
-----------------------------------------------------------------------------------------------

--TODO: Trim Name Lengths

-----------------------------------------------------------------------------------------------
-- PlatesOfAwesome Instance
-----------------------------------------------------------------------------------------------
local PlatesOfAwesomeInst = PlatesOfAwesome:new()
PlatesOfAwesomeInst:Init()
